from .archive import BoundedArchive, NonDominatedSolutionListArchive, CrowdingDistanceArchive,  \
    CrowdingDistanceArchiveWithReferencePoint
from .comparator import EqualSolutionsComparator, SolutionAttributeComparator, RankingAndCrowdingDistanceComparator, \
    DominanceComparator
from .density_estimator import CrowdingDistance
from .evaluator import SequentialEvaluator, MapEvaluator
from .observer import ProgressBarObserver, BasicAlgorithmObserver, WriteFrontToFileObserver, VisualizerObserver
from .quality_indicator import HyperVolume
from .ranking import FastNonDominatedRanking

__all__ = [
    'BoundedArchive', 'NonDominatedSolutionListArchive', 'CrowdingDistanceArchive',
    'CrowdingDistanceArchiveWithReferencePoint',
    'EqualSolutionsComparator', 'SolutionAttributeComparator', 'RankingAndCrowdingDistanceComparator',
    'DominanceComparator',
    'CrowdingDistance',
    'SequentialEvaluator', 'MapEvaluator',
    'ProgressBarObserver', 'BasicAlgorithmObserver', 'WriteFrontToFileObserver', 'VisualizerObserver',
    'HyperVolume',
    'FastNonDominatedRanking'
]



from typing import TypeVar, List

from jmetal.algorithm.singleobjective.evolutionaryalgorithm import GenerationalGeneticAlgorithm
from jmetal.component.evaluator import SequentialEvaluator, Evaluator
from jmetal.core.operator import Mutation, Crossover, Selection
from jmetal.core.problem import Problem
from jmetal.operator.selection import RankingAndCrowdingDistanceSelection

S = TypeVar('S')
R = TypeVar(List[S])

"""
.. module:: NSGA-II
   :platform: Unix, Windows
   :synopsis: NSGA-II (Non-dominance Sorting Genetic Algorithm II) implementation.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class NSGAII(GenerationalGeneticAlgorithm[S, R]):

    def __init__(self,
                 problem: Problem[S],
                 population_size: int,
                 max_evaluations: int,
                 mutation: Mutation[S],
                 crossover: Crossover[S, S],
                 selection: Selection[List[S], S],
                 evaluator: Evaluator[S] = SequentialEvaluator[S]()):
        """  NSGA-II implementation as described in

        * K. Deb, A. Pratap, S. Agarwal and T. Meyarivan, "A fast and elitist
          multiobjective genetic algorithm: NSGA-II," in IEEE Transactions on Evolutionary Computation,
          vol. 6, no. 2, pp. 182-197, Apr 2002. doi: 10.1109/4235.996017

        NSGA-II is a genetic algorithm (GA), i.e. it belongs to the evolutionary algorithms (EAs)
        family. The implementation of NSGA-II provided in jMetalPy follows the evolutionary
        algorithm template described in the algorithm module (:py:mod:`jmetal.core.algorithm`).

        :param problem: The problem to solve.
        :param population_size: Size of the population.
        :param max_evaluations: Maximum number of evaluations/iterations.
        :param mutation: Mutation operator (see :py:mod:`jmetal.operator.mutation`).
        :param crossover: Crossover operator (see :py:mod:`jmetal.operator.crossover`).
        :param selection: Selection operator (see :py:mod:`jmetal.operator.selection`).
        :param evaluator: An evaluator object to evaluate the individuals of the population.
        """
        super(NSGAII, self).__init__(
            problem,
            population_size,
            max_evaluations,
            mutation,
            crossover,
            selection,
            evaluator)

    def replacement(self, population: List[S], offspring_population: List[S]) -> List[List[S]]:
        """ This method joins the current and offspring populations to produce the population of the next generation
        by applying the ranking and crowding distance selection.

        :param population: Parent population.
        :param offspring_population: Offspring population.
        :return: New population after ranking and crowding distance selection is applied.
        """
        join_population = population + offspring_population
        return RankingAndCrowdingDistanceSelection(self.population_size).execute(join_population)

    def get_result(self) -> R:
        return self.population

    def get_name(self) -> str:
        return 'Non-dominated Sorting Genetic Algorithm II'



from abc import ABCMeta
from typing import List, Generic, TypeVar

BitSet = List[bool]
S = TypeVar('S')


class Solution(Generic[S]):
    """ Class representing solutions """

    __metaclass__ = ABCMeta

    def __init__(self, number_of_variables: int, number_of_objectives: int, number_of_constraints: int = 0):
        self.number_of_objectives = number_of_objectives
        self.number_of_variables = number_of_variables
        self.number_of_constraints = number_of_constraints

        self.objectives = [0.0 for _ in range(self.number_of_objectives)]
        self.variables = [[] for _ in range(self.number_of_variables)]
        self.attributes = {}

    def __str__(self) -> str:
        solution = 'number_of_objectives: {0} \nnumber_of_variables: {1} \nnumber_of_constraints: {2} \n'.format(
            self.number_of_objectives, self.number_of_variables, self.number_of_constraints
        )
        solution += 'objectives: \n {0} \n'.format(self.objectives)
        solution += 'variables: \n {0}'.format(self.variables)

        return solution


class BinarySolution(Solution[BitSet]):
    """ Class representing float solutions """

    def __init__(self, number_of_variables: int, number_of_objectives: int, number_of_constraints: int=0):
        super(BinarySolution, self).__init__(number_of_variables, number_of_objectives, number_of_constraints)

    def __copy__(self):
        new_solution = BinarySolution(
            self.number_of_variables,
            self.number_of_objectives,
            self.number_of_constraints)
        new_solution.objectives = self.objectives[:]
        new_solution.variables = self.variables[:]

        return new_solution

    def get_total_number_of_bits(self) -> int:
        total = 0
        for var in self.variables:
            total += len(var)

        return total


class FloatSolution(Solution[float]):
    """ Class representing float solutions """

    def __init__(self, number_of_variables: int, number_of_objectives: int, number_of_constraints: int,
                 lower_bound: List[float], upper_bound: List[float]):
        super(FloatSolution, self).__init__(number_of_variables, number_of_objectives, number_of_constraints)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __copy__(self):
        new_solution = FloatSolution(
            self.number_of_variables,
            self.number_of_objectives,
            self.number_of_constraints,
            self.lower_bound,
            self.upper_bound)
        new_solution.objectives = self.objectives[:]
        new_solution.variables = self.variables[:]

        return new_solution


class IntegerSolution(Solution[int]):
    """ Class representing integer solutions """

    def __init__(self, number_of_variables: int, number_of_objectives: int, number_of_constraints: int,
                 lower_bound: List[int], upper_bound: List[int]):
        super(IntegerSolution, self).__init__(number_of_variables, number_of_objectives, number_of_constraints)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __copy__(self):
        new_solution = FloatSolution(
            self.number_of_variables,
            self.number_of_objectives,
            self.number_of_constraints,
            self.lower_bound,
            self.upper_bound)
        new_solution.objectives = self.objectives[:]
        new_solution.variables = self.variables[:]

        return new_solution



import logging
import os

from tqdm import tqdm

from jmetal.util.graphic import ScatterStreaming
from jmetal.core.observable import Observer
from jmetal.util.solution_list_output import SolutionList

jMetalPyLogger = logging.getLogger('jMetalPy')

"""
.. module:: observer
   :platform: Unix, Windows
   :synopsis: Implementation of algorithm's observers.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class ProgressBarObserver(Observer):

    def __init__(self, step: int, maximum: int, desc: str= 'Progress') -> None:
        """ Show a smart progress meter with the number of evaluations and computing time.

        :param step: Initial counter value.
        :param maximum: Number of expected iterations.
        :param desc: Prefix for the progressbar. """
        self.progress_bar = tqdm(total=maximum, initial=step, ascii=True, desc=desc)
        self.progress = step
        self.step = step
        self.maxx = maximum

    def update(self, *args, **kwargs):
        self.progress_bar.update(self.step)
        self.progress += self.step

        if self.progress >= self.maxx:
            self.progress_bar.close()


class BasicAlgorithmObserver(Observer):

    def __init__(self, frequency: float = 1.0) -> None:
        """ Show the number of evaluations, best fitness and computing time.

        :param frequency: Display frequency. """
        self.display_frequency = frequency

    def update(self, *args, **kwargs):
        computing_time = kwargs['computing time']
        evaluations = kwargs['evaluations']
        front = kwargs['population']

        if (evaluations % self.display_frequency) == 0:
            jMetalPyLogger.debug(
                'Evaluations: {0} \n Best fitness: {1} \n Computing time: {2}'.format(
                    evaluations, front[0].objectives, computing_time
                )
            )


class WriteFrontToFileObserver(Observer):

    def __init__(self, output_directory) -> None:
        """ Write function values of the front into files.

        :param output_directory: Output directory. Each front will be saved on a file `FUN.x`. """
        self.counter = 0
        self.directory = output_directory

        if os.path.isdir(self.directory):
            jMetalPyLogger.warning('Directory {} exists. Removing contents.'.format(self.directory))
            for file in os.listdir(self.directory):
                os.remove('{0}/{1}'.format(self.directory, file))
        else:
            jMetalPyLogger.warning('Directory {} does not exist. Creating it.'.format(self.directory))
            os.mkdir(self.directory)

    def update(self, *args, **kwargs):
        front = kwargs['population']

        SolutionList.print_function_values_to_file(front, '{0}/FUN.{1}'.format(self.directory, self.counter))
        self.counter += 1


class VisualizerObserver(Observer):

    def __init__(self, replace: bool=True) -> None:
        self.display_frequency = 1.0
        self.replace = replace
        self.plot = ScatterStreaming(plot_title='jMetalPy')

    def update(self, *args, **kwargs):
        computing_time = kwargs['computing time']
        evaluations = kwargs['evaluations']

        front = kwargs['population']
        reference_front = kwargs['reference_front']

        title = '{0}, Eval: {1}, Time: {2}'.format('VisualizerObserver', evaluations, computing_time)

        if (evaluations % self.display_frequency) == 0:
            self.plot.update(front, reference_front, rename_title=title, persistence=self.replace)



from typing import TypeVar, List, Generic

from jmetal.component.archive import NonDominatedSolutionListArchive
from jmetal.core.problem import Problem

S = TypeVar('S')

"""
.. module:: RamdomSearch
   :platform: Unix, Windows
   :synopsis: Simple random search algorithms.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class RandomSearch(Generic[S]):

    def __init__(self,
                 problem: Problem[S],
                 max_evaluations: int = 25000):
        self.problem = problem
        self.max_evaluations = max_evaluations
        self.archive = NonDominatedSolutionListArchive()

    def run(self) -> None:
        for i in range(self.max_evaluations):
            new_solution = self.problem.create_solution()
            self.problem.evaluate(new_solution)
            self.archive.add(new_solution)

    def get_result(self) -> List[S]:
        return self.archive.solution_list

    @staticmethod
    def get_name() -> str:
        return 'Random Search Algorithm'



from .multiobjective.nsgaii import NSGAII
from .multiobjective.smpso import SMPSO, SMPSORP
from .singleobjective.evolutionaryalgorithm import ElitistEvolutionStrategy, NonElitistEvolutionStrategy
from .multiobjective.randomSearch import RandomSearch

__all__ = [
    'NSGAII',
    'SMPSO', 'SMPSORP',
    'ElitistEvolutionStrategy', 'NonElitistEvolutionStrategy','RandomSearch'
]



from abc import ABCMeta, abstractmethod
from typing import Generic, TypeVar, List
from pathlib import Path
import random

from jmetal.core.solution import BinarySolution, FloatSolution, IntegerSolution

S = TypeVar('S')


class Problem(Generic[S]):
    """ Class representing problems. """

    __metaclass__ = ABCMeta

    MINIMIZE = -1
    MAXIMIZE = 1

    def __init__(self, reference_front_path: str):
        self.number_of_variables: int = None
        self.number_of_objectives: int = None
        self.number_of_constraints: int = None

        self.obj_directions: List[int] = []
        self.obj_labels: List[str] = []

        self.reference_front: List[S] = None
        if reference_front_path:
            self.reference_front = self.read_front_from_file_as_solutions(reference_front_path)

    @staticmethod
    def read_front_from_file(file_path: str) -> List[List[float]]:
        """ Reads a front from a file and returns a list.

        :return: List of solution points. """
        front = []
        if Path(file_path).is_file():
            with open(file_path) as file:
                for line in file:
                    vector = [float(x) for x in line.split()]
                    front.append(vector)
        else:
            raise Exception('Reference front file was not found at {}'.format(file_path))

        return front

    @staticmethod
    def read_front_from_file_as_solutions(file_path: str) -> List[S]:
        """ Reads a front from a file and returns a list of solution objects.

        :return: List of solution objects. """
        front = []
        if Path(file_path).is_file():
            with open(file_path) as file:
                for line in file:
                    vector = [float(x) for x in line.split()]
                    solution = FloatSolution(2, 2, 0, [], [])
                    solution.objectives = vector

                    front.append(solution)
        else:
            raise Exception('Reference front file was not found at {}'.format(file_path))

        return front

    @abstractmethod
    def create_solution(self) -> S:
        """ Creates a random solution to the problem.

        :return: Solution. """
        pass

    @abstractmethod
    def evaluate(self, solution: S) -> S:
        """ Evaluate a solution. For any new problem inheriting from :class:`Problem`, this method should be replaced.

        :return: Evaluated solution. """
        pass

    def evaluate_constraints(self, solution: S):
        pass

    def get_name(self) -> str:
        return self.__class__.__name__


class BinaryProblem(Problem[BinarySolution]):
    """ Class representing binary problems. """

    __metaclass__ = ABCMeta

    def __init__(self, rf_path: str = None):
        super(BinaryProblem, self).__init__(reference_front_path=rf_path)

    def create_solution(self) -> BinarySolution:
        pass

    @abstractmethod
    def evaluate(self, solution: BinarySolution) -> BinarySolution:
        pass


class FloatProblem(Problem[FloatSolution]):
    """ Class representing float problems. """

    __metaclass__ = ABCMeta

    def __init__(self, rf_path: str = None):
        super(FloatProblem, self).__init__(reference_front_path=rf_path)
        self.lower_bound = None
        self.upper_bound = None

    def create_solution(self) -> FloatSolution:
        new_solution = FloatSolution(self.number_of_variables, self.number_of_objectives, self.number_of_constraints,
                                     self.lower_bound, self.upper_bound)
        new_solution.variables = \
            [random.uniform(self.lower_bound[i]*1.0, self.upper_bound[i]*1.0) for i in range(self.number_of_variables)]

        return new_solution

    @abstractmethod
    def evaluate(self, solution: FloatSolution) -> FloatSolution:
        pass


class IntegerProblem(Problem[IntegerSolution]):
    """ Class representing integer problems. """

    __metaclass__ = ABCMeta

    def __init__(self, rf_path: str = None):
        super(IntegerProblem, self).__init__(reference_front_path=rf_path)
        self.lower_bound = None
        self.upper_bound = None

    def create_solution(self) -> IntegerSolution:
        new_solution = IntegerSolution(
            self.number_of_variables,
            self.number_of_objectives,
            self.number_of_constraints,
            self.lower_bound, self.upper_bound)

        new_solution.variables = \
            [int(random.uniform(self.lower_bound[i]*1.0, self.upper_bound[i]*1.0))
             for i in range(self.number_of_variables)]

        return new_solution

    @abstractmethod
    def evaluate(self, solution: IntegerSolution) -> IntegerSolution:
        pass



import threading
import time
from abc import ABCMeta, abstractmethod
from typing import TypeVar, Generic, List

from jmetal.core.solution import FloatSolution
from jmetal.core.observable import DefaultObservable

S = TypeVar('S')
R = TypeVar('R')

"""
.. module:: algorithm
   :platform: Unix, Windows
   :synopsis: Templates for algorithms.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class Algorithm(Generic[S, R], threading.Thread):

    __metaclass__ = ABCMeta

    def __init__(self):
        threading.Thread.__init__(self)
        self.evaluations = 0
        self.max_evaluations = 100
        self.start_computing_time = 0
        self.total_computing_time = 0
        self.observable = DefaultObservable()

    def get_evaluations(self) -> int:
        return self.evaluations

    def get_current_computing_time(self) -> float:
        return time.time() - self.start_computing_time

    @abstractmethod
    def get_result(self) -> R:
        """ :return: Final population.
        """
        pass

    def get_name(self) -> str:
        return self.__class__.__name__


class EvolutionaryAlgorithm(Algorithm[S, R]):

    __metaclass__ = ABCMeta

    def __init__(self):
        super(EvolutionaryAlgorithm, self).__init__()
        self.population = []

    @abstractmethod
    def create_initial_population(self) -> List[S]:
        pass

    @abstractmethod
    def evaluate_population(self, population: List[S]) -> List[S]:
        pass

    @abstractmethod
    def init_progress(self) -> None:
        pass

    @abstractmethod
    def is_stopping_condition_reached(self) -> bool:
        pass

    @abstractmethod
    def selection(self, population: List[S]) -> List[S]:
        pass

    @abstractmethod
    def reproduction(self, population: List[S]) -> List[S]:
        pass

    @abstractmethod
    def replacement(self, population: List[S], offspring_population: List[S]) -> List[S]:
        pass

    @abstractmethod
    def update_progress(self):
        pass

    def run(self):
        """* Step One: Generate the initial population of individuals randomly. (First generation)
        * Step Two: Evaluate the fitness of each individual in that population
        * Step Three: Repeat the following regenerational steps until termination

            1. Select the best-fit individuals for reproduction. (Parents)
            2. Breed new individuals through crossover and mutation operations to give birth to offspring.
            3. Evaluate the individual fitness of new individuals.
            4. Replace least-fit population with new individuals.

        .. note::
            To develop an EA, all the abstract the methods used in the run() method must be implemented.
        """

        self.start_computing_time = time.time()

        self.population = self.create_initial_population()
        self.population = self.evaluate_population(self.population)
        self.init_progress()

        while not self.is_stopping_condition_reached():
            mating_population = self.selection(self.population)
            offspring_population = self.reproduction(mating_population)
            offspring_population = self.evaluate_population(offspring_population)
            self.population = self.replacement(self.population, offspring_population)
            self.update_progress()

        self.total_computing_time = self.get_current_computing_time()


class ParticleSwarmOptimization(Algorithm[FloatSolution, List[FloatSolution]]):

    __metaclass__ = ABCMeta

    def __init__(self):
        super(ParticleSwarmOptimization, self).__init__()
        self.swarm = []

    @abstractmethod
    def init_progress(self) -> None:
        pass

    @abstractmethod
    def update_progress(self) -> None:
        pass

    @abstractmethod
    def is_stopping_condition_reached(self) -> bool:
        pass

    @abstractmethod
    def create_initial_swarm(self) -> List[FloatSolution]:
        pass

    @abstractmethod
    def evaluate_swarm(self, swarm: List[FloatSolution]) -> List[FloatSolution]:
        pass

    @abstractmethod
    def initialize_global_best(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def initialize_particle_best(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def initialize_velocity(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def update_velocity(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def update_position(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def perturbation(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def update_global_best(self, swarm: List[FloatSolution]) -> None:
        pass

    @abstractmethod
    def update_particle_best(self, swarm: List[FloatSolution]) -> None:
        pass

    def run(self):
        self.start_computing_time = time.time()

        self.swarm = self.create_initial_swarm()
        self.swarm = self.evaluate_swarm(self.swarm)
        self.initialize_velocity(self.swarm)
        self.initialize_particle_best(self.swarm)
        self.initialize_global_best(self.swarm)
        self.init_progress()

        while not self.is_stopping_condition_reached():
            self.update_velocity(self.swarm)
            self.update_position(self.swarm)
            self.perturbation(self.swarm)
            self.swarm = self.evaluate_swarm(self.swarm)
            self.update_global_best(self.swarm)
            self.update_particle_best(self.swarm)
            self.update_progress()

        self.total_computing_time = self.get_current_computing_time()



from abc import ABCMeta, abstractmethod
from typing import TypeVar, Generic

from jmetal.core.solution import Solution

S = TypeVar('S')


class Comparator(Generic[S]):

    __metaclass__ = ABCMeta

    @abstractmethod
    def compare(self, solution1: S, solution2: S) -> int:
        pass

    def get_name(self) -> str:
        return self.__class__.__name__


class EqualSolutionsComparator(Comparator):

    def compare(self, solution1: Solution, solution2: Solution) -> int:
        if solution1 is None:
            return 1
        elif solution2 is None:
            return -1

        dominate1 = 0
        dominate2 = 0

        for i in range(len(solution1.objectives)):
            value1 = solution1.objectives[i]
            value2 = solution2.objectives[i]

            if value1 < value2:
                flag = -1
            elif value1 > value2:
                flag = 1
            else:
                flag = 0

            if flag == -1:
                dominate1 = 1
            if flag == 1:
                dominate2 = 1

        if dominate1 == 0 and dominate2 == 0:
            return 0
        elif dominate1 == 1:
            return -1
        elif dominate2 == 1:
            return 1


class SolutionAttributeComparator(Comparator):

    def __init__(self, key: str, lowest_is_best: bool = True):
        self.key = key
        self.lowest_is_best = lowest_is_best

    def compare(self, solution1: Solution, solution2: Solution) -> int:
        value1 = solution1.attributes.get(self.key)
        value2 = solution2.attributes.get(self.key)

        result = 0
        if value1 is not None and value2 is not None:
            if self.lowest_is_best:
                if value1 < value2:
                    result = -1
                elif value1 > value2:
                    result = 1
                else:
                    result = 0
            else:
                if value1 > value2:
                    result = -1
                elif value1 < value2:
                    result = 1
                else:
                    result = 0

        return result


class RankingAndCrowdingDistanceComparator(Comparator):

    def compare(self, solution1: Solution, solution2: Solution) -> int:
        result = \
            SolutionAttributeComparator("dominance_ranking").compare(solution1, solution2)

        if result is 0:
            result = \
                SolutionAttributeComparator("crowding_distance", lowest_is_best=False).compare(solution1, solution2)

        return result


class DominanceComparator(Comparator):

    def __init__(self, constraint_comparator=SolutionAttributeComparator("overall_constraint_violation", False)):
        self.constraint_comparator = constraint_comparator

    def compare(self, solution1: Solution, solution2: Solution) -> int:
        if solution1 is None:
            raise Exception("The solution1 is None")
        elif solution2 is None:
            raise Exception("The solution2 is None")
        # elif len(solution1.objectives) != len(solution2.objectives):
        #    raise Exception("The solutions have different number of objectives")

        result = self.constraint_comparator.compare(solution1, solution2)
        if result == 0:
            result = self.__dominance_test(solution1, solution2)

        return result

    def __dominance_test(self, solution1: Solution, solution2: Solution) -> float:
        best_is_one = 0
        best_is_two = 0

        for i in range(solution1.number_of_objectives):
            value1 = solution1.objectives[i]
            value2 = solution2.objectives[i]
            if value1 != value2:
                if value1 < value2:
                    best_is_one = 1
                if value1 > value2:
                    best_is_two = 1

        if best_is_one > best_is_two:
            result = -1
        elif best_is_two > best_is_one:
            result = 1
        else:
            result = 0

        return result



import logging
from abc import ABCMeta, abstractmethod
from typing import TypeVar, List

logger = logging.getLogger(__name__)

S = TypeVar('S')

"""
.. module:: crowding_distance
   :platform: Unix, Windows
   :synopsis: Crowding distance implementation.

.. moduleauthor:: Álvaro Gómez Jáuregui <alvarogj@lcc.uma.es>
"""


class DensityEstimator(List[S]):
    """This is the interface of any density estimator algorithm.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def compute_density_estimator(self, solution_list: List[S]) -> float:
        pass


class CrowdingDistance(DensityEstimator[List[S]]):
    """This class implements a DensityEstimator based on the crowding distance.
    In consequence, the main method of this class is :func:`compute_density_estimator`.
    """

    def compute_density_estimator(self, front: List[S]):
        """This function performs the computation of the crowding density estimation over the solution list.

        .. note::
           This method assign the distance in the inner elements of the solution list.

        :param front: The list of solutions.
        """
        size = len(front)

        if size is 0:
            return
        elif size is 1:
            front[0].attributes['crowding_distance'] = float("inf")
            return
        elif size is 2:
            front[0].attributes['crowding_distance'] = float("inf")
            front[1].attributes['crowding_distance'] = float("inf")
            return

        for i in range(len(front)):
            front[i].attributes['crowding_distance'] = 0.0

        number_of_objectives = front[0].number_of_objectives

        for i in range(number_of_objectives):
            # Sort the population by Obj n
            front = sorted(front, key=lambda x: x.objectives[i])
            objective_minn = front[0].objectives[i]
            objective_maxn = front[len(front) - 1].objectives[i]

            # Set de crowding distance
            front[0].attributes['crowding_distance'] = float('inf')
            front[size - 1].attributes['crowding_distance'] = float('inf')

            for j in range(1, size - 1):
                distance = front[j + 1].objectives[i] - front[j - 1].objectives[i]

                # Check if minimum and maximum are the same (in which case do nothing)
                if objective_maxn - objective_minn == 0:
                    logger.warning('Minimum and maximum are the same!')
                else:
                    distance = distance / (objective_maxn - objective_minn)

                distance += front[j].attributes['crowding_distance']
                front[j].attributes['crowding_distance'] = distance



import logging
from typing import TypeVar, Generic

logger = logging.getLogger(__name__)

S = TypeVar('S')

"""
.. module:: solution_list
   :platform: Unix, Windows
   :synopsis: Utils to print solutions.

.. moduleauthor:: Antonio J. Nebro <ajnebro@uma.es>
"""


class SolutionList(Generic[S]):

    @staticmethod
    def print_variables_to_screen(solution_list: list):
        for solution in solution_list:
            print(solution.variables[0])

    @staticmethod
    def print_variables_to_file(solution_list: list, file_name):
        logger.info("Output file (variables): " + file_name)
        with open(file_name, 'w') as of:
            for solution in solution_list:
                for variables in solution.variables:
                    of.write(str(variables) + " ")
                of.write("\n")

    @staticmethod
    def print_function_values_to_screen(solution_list: list):
        for solution in solution_list:
            print(str(solution_list.index(solution)) + ": ", sep='  ', end='', flush=True)
            print(solution.objectives, sep='  ', end='', flush=True)
            print()

    @staticmethod
    def print_function_values_to_file(solution_list: list, file_name):
        logger.info("Output file (function values): " + file_name)
        with open(file_name, 'w') as of:
            for solution in solution_list:
                for function_value in solution.objectives:
                    of.write(str(function_value) + " ")
                of.write("\n")



import random
import copy
from abc import ABCMeta, abstractmethod
from typing import TypeVar, Generic, List

from jmetal.component.density_estimator import CrowdingDistance, DensityEstimator
from jmetal.component.comparator import Comparator, DominanceComparator, EqualSolutionsComparator, \
    SolutionAttributeComparator

S = TypeVar('S')

"""
.. module:: archive
   :platform: Unix, Windows
   :synopsis: Archive implementation.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class Archive(Generic[S]):

    __metaclass__ = ABCMeta

    def __init__(self):
        self.solution_list: List[S] = []

    @abstractmethod
    def add(self, solution: S) -> bool:
        pass

    def get(self, index: int) -> S:
        return self.solution_list[index]

    def size(self) -> int:
        return len(self.solution_list)

    def get_name(self) -> str:
        return self.__class__.__name__


class BoundedArchive(Archive[S]):

    def __init__(self,
                 maximum_size: int,
                 comparator: Comparator[S]=None,
                 density_estimator: DensityEstimator=None):
        super(BoundedArchive, self).__init__()
        self.maximum_size = maximum_size
        self.comparator = comparator
        self.density_estimator = density_estimator
        self.non_dominated_solution_archive = NonDominatedSolutionListArchive()
        self.solution_list = self.non_dominated_solution_archive.solution_list

    def compute_density_estimator(self):
        self.density_estimator.compute_density_estimator(self.solution_list)

    def add(self, solution: S) -> bool:
        success = self.non_dominated_solution_archive.add(solution)
        if success:
            if self.size() > self.maximum_size:
                self.compute_density_estimator()
                worst_solution = self.__find_worst_solution(self.solution_list)
                self.solution_list.remove(worst_solution)

        return success

    def __find_worst_solution(self, solution_list: List[S]) -> S:
        if solution_list is None:
            raise Exception("The solution list is None")
        elif len(solution_list) is 0:
            raise Exception("The solution list is empty")

        worst_solution = solution_list[0]
        for solution in solution_list[1:]:
            if self.comparator.compare(worst_solution, solution) < 0:
                worst_solution = solution

        return worst_solution


class NonDominatedSolutionListArchive(Archive[S]):

    def __init__(self):
        super(NonDominatedSolutionListArchive, self).__init__()
        self.comparator = DominanceComparator()

    def add(self, solution: S) -> bool:
        is_dominated = False
        is_contained = False

        if len(self.solution_list) == 0:
            self.solution_list.append(solution)
            return True
        else:
            number_of_deleted_solutions = 0

            # New copy of list and enumerate
            for index, current_solution in enumerate(list(self.solution_list)):
                is_dominated_flag = self.comparator.compare(solution, current_solution)
                if is_dominated_flag == -1:
                    del self.solution_list[index-number_of_deleted_solutions]
                    number_of_deleted_solutions += 1
                elif is_dominated_flag == 1:
                    is_dominated = True
                    break
                elif is_dominated_flag == 0:
                    if EqualSolutionsComparator().compare(solution, current_solution) == 0:
                        is_contained = True
                        break

        if not is_dominated and not is_contained:
            self.solution_list.append(solution)
            return True

        return False


class CrowdingDistanceArchive(BoundedArchive[S]):

    def __init__(self,
                 maximum_size: int):
        super(CrowdingDistanceArchive, self).__init__(
            maximum_size=maximum_size,
            comparator=SolutionAttributeComparator("crowding_distance", lowest_is_best=False),
            density_estimator=CrowdingDistance())


class ArchiveWithReferencePoint(BoundedArchive[S]):

    def __init__(self,
                 maximum_size: int,
                 reference_point: List[float],
                 comparator: Comparator[S],
                 density_estimator: DensityEstimator):
        super(ArchiveWithReferencePoint, self).__init__(maximum_size, comparator, density_estimator)
        self.__reference_point = reference_point
        self.__comparator = comparator
        self.__density_estimator = density_estimator
        self.__reference_point_solution = None

    def add(self, solution: S) -> bool:
        if self.__reference_point_solution is None:
            self.__reference_point_solution = copy.deepcopy(solution)

        self.__reference_point_solution.objectives = [value for value in self.__reference_point]

        dominated_solution = None

        if self.__dominance_test(solution, self.__reference_point_solution) == 0:
            if len(self.solution_list) == 0:
                result = True
            else:
                if random.uniform(0.0, 1.0) < 0.05:
                    result = True
                    dominated_solution = solution
                else:
                    result = False
        else:
            result = True

        if result:
            result = super(ArchiveWithReferencePoint, self).add(solution)

        if result and dominated_solution is not None and len(self.solution_list) > 1:
            self.solution_list.remove(dominated_solution)

        if result and len(self.solution_list) > self.maximum_size:
            self.compute_density_estimator()

        return result

    def get_reference_point(self)->List[float]:
        return self.__reference_point

    def __dominance_test(self, solution1: S, solution2: S) -> int:
        best_is_one = 0
        best_is_two = 0

        for value1, value2 in zip(solution1.objectives, solution2.objectives):
            if value1 != value2:
                if value1 < value2:
                    best_is_one = 1
                if value2 < value1:
                    best_is_two = 1

        if best_is_one > best_is_two:
            result = -1
        elif best_is_two > best_is_one:
            result = 1
        else:
            result = 0

        return result


class CrowdingDistanceArchiveWithReferencePoint(ArchiveWithReferencePoint[S]):

    def __init__(self,
                 maximum_size: int,
                 reference_point: List[float]):
        super(CrowdingDistanceArchiveWithReferencePoint, self).__init__(
            maximum_size=maximum_size,
            reference_point=reference_point,
            comparator=SolutionAttributeComparator("crowding_distance", lowest_is_best=False),
            density_estimator=CrowdingDistance())



import logging
from concurrent.futures import ProcessPoolExecutor

jMetalPyLogger = logging.getLogger('jMetalPy')

"""
.. module:: laboratory
   :platform: Unix, Windows
   :synopsis: Run experiments. WIP!

.. moduleauthor:: Antonio Benítez-Hidalgo <antonio.b@uma.es>
"""


class Experiment:

    def __init__(self, algorithm_list: list, n_runs: int = 1, m_workers: int = 6):
        """ :param algorithm_list: List of algorithms as Tuple(Algorithm, dic() with parameters).
        :param m_workers: Maximum number of workers for ProcessPoolExecutor. """
        self.algorithm_list = algorithm_list

        self.n_runs = n_runs
        self.m_workers = m_workers
        self.experiment_list = list()

    def run(self) -> None:
        """ Run the experiment. """
        self.__configure_algorithm_list()

        with ProcessPoolExecutor(max_workers=self.m_workers) as pool:
            for name, algorithm, n_run in self.experiment_list:
                jMetalPyLogger.info('Running experiment {0}'.format(
                    name, algorithm.problem, n_run)
                )

                pool.submit(algorithm.run())

            # Wait until all computation is done for this problem
            jMetalPyLogger.debug('Waiting')
            pool.shutdown(wait=True)

    def compute_metrics(self, metric_list: list) -> dict:
        """ :param metric_list: List of metrics. Each metric should inherit from :py:class:`Metric` or, at least,
        contain a method `compute`. """
        results = dict()

        for name, algorithm, n_run in self.experiment_list:
            results[name] = {}

            for metric in metric_list:
                results[name].setdefault('metric', dict()).update(
                    {metric.get_name(): metric.compute(algorithm.get_result())}
                )

        return results

    def export_to_file(self, base_directory: str = 'experiment', function_values_filename: str = 'FUN',
                       variables_filename: str = 'VAR'):
        for tag, algorithm, run in self.experiment_list:
            # todo Save VAR and FUN to files
            pass

    def __configure_algorithm_list(self):
        """ Configure the algorithm list, by making a triple of (name, algorithm, run). """
        for n_run in range(self.n_runs):
            for tag, algorithm in self.algorithm_list:
                name = '{0}.{1}.{2}'.format(tag, algorithm.problem.get_name(), n_run)
                self.experiment_list.append((name, algorithm, n_run))



from abc import ABCMeta, abstractmethod

"""
.. module:: Observable
   :platform: Unix, Windows
   :synopsis: Implementation of the observer-observable pattern.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class Observer:

    __metaclass__ = ABCMeta

    @abstractmethod
    def update(self, *args, **kwargs):
        """ Update method

        :param args:
        :param kwargs:
        :return:
        """
        pass


class Observable:

    __metaclass__ = ABCMeta

    @abstractmethod
    def register(self, observer):
        pass

    @abstractmethod
    def deregister(self, observer):
        pass

    @abstractmethod
    def deregister_all(self):
        pass

    @abstractmethod
    def notify_all(self, *args, **kwargs):
        pass


class DefaultObservable(Observable):

    def __init__(self):
        self.observers = []

    def register(self, observer: Observer):
        if observer not in self.observers:
            self.observers.append(observer)

    def deregister(self, observer: Observer):
        if observer in self.observers:
            self.observers.remove(observer)

    def deregister_all(self):
        if self.observers:
            del self.observers[:]

    def notify_all(self, *args, **kwargs):
        for observer in self.observers:
            observer.update(*args, **kwargs)



import random
from typing import List, TypeVar

from jmetal.component.density_estimator import CrowdingDistance
from jmetal.core.operator import Selection
from jmetal.component.comparator import Comparator, DominanceComparator
from jmetal.component.ranking import FastNonDominatedRanking

S = TypeVar('S')

"""
.. module:: selection
   :platform: Unix, Windows
   :synopsis: Module implementing selection operators.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class BinaryTournamentSelection(Selection[List[S], S]):

    def __init__(self, comparator: Comparator = DominanceComparator()):
        super(BinaryTournamentSelection, self).__init__()
        self.comparator = comparator

    def execute(self, front: List[S]) -> S:
        if front is None:
            raise Exception('The front is null')
        elif len(front) == 0:
            raise Exception('The front is empty')

        if len(front) == 1:
            result = front[0]
        else:
            # Sampling without replacement
            i, j = random.sample(range(0, len(front)), 2)
            solution1 = front[i]
            solution2 = front[j]

            flag = self.comparator.compare(solution1, solution2)

            if flag == -1:
                result = solution1
            elif flag == 1:
                result = solution2
            else:
                result = [solution1, solution2][random.random() < 0.5]

        return result

    def get_name(self) -> str:
        return 'Binary tournament selection'


class BestSolutionSelection(Selection[List[S], S]):

    def __init__(self):
        super(BestSolutionSelection, self).__init__()

    def execute(self, front: List[S]) -> S:
        if front is None:
            raise Exception('The front is null')
        elif len(front) == 0:
            raise Exception('The front is empty')

        result = front[0]
        for solution in front[1:]:
            if DominanceComparator().compare(solution, result) < 0:
                result = solution

        return result

    def get_name(self) -> str:
        return 'Best solution selection'


class NaryRandomSolutionSelection(Selection[List[S], S]):

    def __init__(self, number_of_solutions_to_be_returned: int = 1):
        super(NaryRandomSolutionSelection, self).__init__()
        if number_of_solutions_to_be_returned < 0:
            raise Exception('The number of solutions to be returned must be positive integer')

        self.number_of_solutions_to_be_returned = number_of_solutions_to_be_returned

    def execute(self, front: List[S]) -> S:
        if front is None:
            raise Exception('The front is null')
        if len(front) == 0:
            raise Exception('The front is empty')
        if len(front) < self.number_of_solutions_to_be_returned:
            raise Exception('The front contains less elements than required')

        # random sampling without replacement
        return random.sample(front, self.number_of_solutions_to_be_returned)

    def get_name(self) -> str:
        return 'Nary random solution selection'


class RandomSolutionSelection(Selection[List[S], S]):

    def __init__(self):
        super(RandomSolutionSelection, self).__init__()

    def execute(self, front: List[S]) -> S:
        if front is None:
            raise Exception('The front is null')
        elif len(front) == 0:
            raise Exception('The front is empty')

        return random.choice(front)

    def get_name(self) -> str:
        return 'Random solution selection'


class RankingAndCrowdingDistanceSelection(Selection[List[S], List[S]]):

    def __init__(self, max_population_size: int):
        super(RankingAndCrowdingDistanceSelection, self).__init__()
        self.max_population_size = max_population_size

    def execute(self, front: List[S]) -> List[S]:
        ranking = FastNonDominatedRanking()
        crowding_distance = CrowdingDistance()
        ranking.compute_ranking(front)

        ranking_index = 0
        new_solution_list = []

        while len(new_solution_list) < self.max_population_size:
            if len(ranking.get_subfront(ranking_index)) < self.max_population_size - len(new_solution_list):
                new_solution_list = new_solution_list + ranking.get_subfront(ranking_index)
                ranking_index += 1
            else:
                subfront = ranking.get_subfront(ranking_index)
                crowding_distance.compute_density_estimator(subfront)
                sorted_subfront = sorted(subfront, key=lambda x: x.attributes['crowding_distance'], reverse=True)
                for i in range((self.max_population_size - len(new_solution_list))):
                    new_solution_list.append(sorted_subfront[i])

        return new_solution_list

    def get_name(self) -> str:
        return 'Ranking and crowding distance selection'


class BinaryTournament2Selection(Selection[List[S], S]):

    def __init__(self, comparator_list: List[Comparator]):
        super(BinaryTournament2Selection, self).__init__()
        self.comparator_list = comparator_list

    def execute(self, front: List[S]) -> S:
        if front is None:
            raise Exception('The front is null')
        elif len(front) == 0:
            raise Exception('The front is empty')
        elif not self.comparator_list:
            raise Exception('The comparators\' list is empty')

        winner = None

        if len(front) == 1:
            winner = front[0]
        else:
            for comparator in self.comparator_list:
                winner = self.__winner(front, comparator)
                if winner is not None:
                    break

        if not winner:
            i = random.randrange(0, len(front))
            winner = front[i]

        return winner

    def __winner(self, front: List[S], comparator: Comparator):
        # Sampling without replacement
        i, j = random.sample(range(0, len(front)), 2)

        solution1 = front[i]
        solution2 = front[j]

        flag = comparator.compare(solution1, solution2)

        if flag == -1:
            result = solution1
        elif flag == 1:
            result = solution2
        else:
            result = None

        return result

    def get_name(self) -> str:
        return 'Binary tournament selection (experimental)'



from abc import ABCMeta, abstractmethod
from multiprocessing.pool import ThreadPool
from typing import TypeVar, List, Generic

from jmetal.core.problem import Problem

S = TypeVar('S')


class Evaluator(Generic[S]):

    __metaclass__ = ABCMeta

    @abstractmethod
    def evaluate(self, solution_list: List[S], problem: Problem) -> List[S]:
        pass

    @staticmethod
    def evaluate_solution(solution: S, problem: Problem) -> None:
        problem.evaluate(solution)
        if problem.number_of_constraints > 0:
            problem.evaluate_constraints(solution)

    def get_name(self) -> str:
        return self.__class__.__name__


class SequentialEvaluator(Evaluator[S]):

    def evaluate(self, solution_list: List[S], problem: Problem) -> List[S]:
        for solution in solution_list:
            Evaluator.evaluate_solution(solution, problem)

        return solution_list


class MapEvaluator(Evaluator[S]):

    def __init__(self, processes=None):
        self.pool = ThreadPool(processes)

    def evaluate(self, solution_list: List[S], problem: Problem) -> List[S]:
        self.pool.map(lambda solution: Evaluator[S].evaluate_solution(solution, problem), solution_list)

        return solution_list



from typing import TypeVar, List
from copy import copy
from math import sqrt
import random

import numpy

from jmetal.component.archive import BoundedArchive
from jmetal.component.evaluator import Evaluator, SequentialEvaluator
from jmetal.core.algorithm import ParticleSwarmOptimization
from jmetal.core.operator import Mutation
from jmetal.core.problem import FloatProblem
from jmetal.core.solution import FloatSolution
from jmetal.component.comparator import DominanceComparator

R = TypeVar('R')

"""
.. module:: SMPSO
   :platform: Unix, Windows
   :synopsis: Implementation of SMPSO.

.. moduleauthor:: Antonio Benítez-Hidalgo <antonio.b@uma.es>
"""


class SMPSO(ParticleSwarmOptimization):

    def __init__(self,
                 problem: FloatProblem,
                 swarm_size: int,
                 max_evaluations: int,
                 mutation: Mutation[FloatSolution],
                 leaders: BoundedArchive[FloatSolution],
                 evaluator: Evaluator[FloatSolution] = SequentialEvaluator[FloatSolution]()):
        """ This class implements the SMPSO algorithm as described in

        * SMPSO: A new PSO-based metaheuristic for multi-objective optimization
        * MCDM 2009. DOI: `<http://dx.doi.org/10.1109/MCDM.2009.4938830/>`_.

        The implementation of SMPSO provided in jMetalPy follows the algorithm template described in the algorithm
        templates section of the documentation.

        :param problem: The problem to solve.
        :param swarm_size: Swarm size.
        :param max_evaluations: Maximum number of evaluations.
        :param mutation: Mutation operator.
        :param leaders: Archive for leaders.
        :param evaluator: An evaluator object to evaluate the solutions in the population.
        """
        super(SMPSO, self).__init__()
        self.problem = problem
        self.swarm_size = swarm_size
        self.max_evaluations = max_evaluations
        self.mutation = mutation
        self.leaders = leaders
        self.evaluator = evaluator

        self.evaluations = 0

        self.c1_min = 1.5
        self.c1_max = 2.5
        self.c2_min = 1.5
        self.c2_max = 2.5
        self.r1_min = 0.0
        self.r1_max = 1.0
        self.r2_min = 0.0
        self.r2_max = 1.0
        self.min_weight = 0.1
        self.max_weight = 0.1

        self.change_velocity1 = -1
        self.change_velocity2 = -1

        self.dominance_comparator = DominanceComparator()

        self.speed = numpy.zeros((self.swarm_size, self.problem.number_of_variables), dtype=float)
        self.delta_max, self.delta_min = numpy.empty(problem.number_of_variables),\
                                         numpy.empty(problem.number_of_variables)
        for i in range(problem.number_of_variables):
            self.delta_max[i] = (self.problem.upper_bound[i] - self.problem.lower_bound[i]) / 2.0

        self.delta_min = -1.0 * self.delta_max

    def init_progress(self) -> None:
        self.evaluations = self.swarm_size
        self.leaders.compute_density_estimator()

    def update_progress(self) -> None:
        self.evaluations += self.swarm_size
        self.leaders.compute_density_estimator()

        observable_data = {'evaluations': self.evaluations,
                           'computing time': self.get_current_computing_time(),
                           'population': self.leaders.solution_list,
                           'reference_front': self.problem.reference_front}

        self.observable.notify_all(**observable_data)

    def is_stopping_condition_reached(self) -> bool:
        return self.evaluations >= self.max_evaluations

    def create_initial_swarm(self) -> List[FloatSolution]:
        swarm = []
        for _ in range(self.swarm_size):
            swarm.append(self.problem.create_solution())
        return swarm

    def evaluate_swarm(self, swarm: List[FloatSolution]) -> List[FloatSolution]:
        return self.evaluator.evaluate(swarm, self.problem)

    def initialize_global_best(self, swarm: List[FloatSolution]) -> None:
        for particle in swarm:
            self.leaders.add(particle)

    def initialize_particle_best(self, swarm: List[FloatSolution]) -> None:
        for particle in swarm:
            particle.attributes['local_best'] = copy(particle)

    def initialize_velocity(self, swarm: List[FloatSolution]) -> None:
        pass  # Velocity initialized in the constructor

    def update_velocity(self, swarm: List[FloatSolution]) -> None:
        for i in range(self.swarm_size):
            best_particle = copy(swarm[i].attributes['local_best'])
            best_global = self.select_global_best()

            r1 = round(random.uniform(self.r1_min, self.r1_max), 1)
            r2 = round(random.uniform(self.r2_min, self.r2_max), 1)
            c1 = round(random.uniform(self.c1_min, self.c1_max), 1)
            c2 = round(random.uniform(self.c2_min, self.c2_max), 1)
            wmax = self.max_weight
            wmin = self.min_weight

            for var in range(swarm[i].number_of_variables):
                self.speed[i][var] = \
                    self.__velocity_constriction(
                        self.__constriction_coefficient(c1, c2) *
                        ((self.__inertia_weight(self.evaluations, self.max_evaluations, wmax, wmin)
                          * self.speed[i][var])
                         + (c1 * r1 * (best_particle.variables[var] - swarm[i].variables[var]))
                         + (c2 * r2 * (best_global.variables[var] - swarm[i].variables[var]))
                         ),
                        self.delta_max, self.delta_min, var)

    def update_position(self, swarm: List[FloatSolution]) -> None:
        for i in range(self.swarm_size):
            particle = swarm[i]

            for j in range(particle.number_of_variables):
                particle.variables[j] += self.speed[i][j]

                if particle.variables[j] < self.problem.lower_bound[j]:
                    particle.variables[j] = self.problem.lower_bound[j]
                    self.speed[i][j] *= self.change_velocity1

                if particle.variables[j] > self.problem.upper_bound[j]:
                    particle.variables[j] = self.problem.upper_bound[j]
                    self.speed[i][j] *= self.change_velocity2

    def perturbation(self, swarm: List[FloatSolution]) -> None:
        for i in range(self.swarm_size):
            if (i % 6) == 0:
                self.mutation.execute(swarm[i])

    def update_global_best(self, swarm: List[FloatSolution]) -> None:
        for particle in swarm:
            self.leaders.add(copy(particle))

    def update_particle_best(self, swarm: List[FloatSolution]) -> None:
        for i in range(self.swarm_size):
            flag = self.dominance_comparator.compare(
                swarm[i],
                swarm[i].attributes['local_best'])
            if flag != 1:
                swarm[i].attributes['local_best'] = copy(swarm[i])

    def get_result(self) -> List[FloatSolution]:
        return self.leaders.solution_list

    def select_global_best(self) -> FloatSolution:
        leaders = self.leaders.solution_list

        if len(leaders) > 2:
            particles = random.sample(leaders, 2)

            if self.leaders.comparator.compare(particles[0], particles[1]) < 1:
                best_global = copy(particles[0])
            else:
                best_global = copy(particles[1])
        else:
            best_global = copy(self.leaders.solution_list[0])

        return best_global

    def __velocity_constriction(self, value: float, delta_max: [], delta_min: [], variable_index: int) -> float:
        result = value
        if value > delta_max[variable_index]:
            result = delta_max[variable_index]
        if value < delta_min[variable_index]:
            result = delta_min[variable_index]

        return result

    def __inertia_weight(self, evaluations: int, max_evaluations: int, wmax: float, wmin: float):
        # todo ?
        return wmax

    def __constriction_coefficient(self, c1: float, c2: float) -> float:
        rho = c1 + c2
        if rho <= 4:
            result = 1.0
        else:
            result = 2.0 / (2.0 - rho - sqrt(pow(rho, 2.0) - 4.0 * rho))

        return result


class SMPSORP(SMPSO):

    def __init__(self,
                 problem: FloatProblem,
                 swarm_size: int,
                 max_evaluations: int,
                 mutation: Mutation[FloatSolution],
                 reference_points: List[List[float]],
                 leaders: List[BoundedArchive[FloatSolution]],
                 evaluator: Evaluator[FloatSolution] = SequentialEvaluator[FloatSolution]()):
        """ This class implements the SMPSORP algorithm.

        :param problem: The problem to solve.
        :param swarm_size:
        :param max_evaluations:
        :param mutation:
        :param leaders: List of bounded archives.
        :param evaluator: An evaluator object to evaluate the solutions in the population.
        """
        super(SMPSORP, self).__init__(
            problem=problem,
            swarm_size=swarm_size,
            max_evaluations=max_evaluations,
            mutation=mutation,
            leaders=None,
            evaluator=evaluator)
        self.reference_points = reference_points
        self.leaders = leaders

    def update_leaders_density_estimator(self):
        for leader in self.leaders:
            leader.compute_density_estimator()

    def init_progress(self) -> None:
        self.evaluations = self.swarm_size
        self.update_leaders_density_estimator()

    def update_progress(self) -> None:
        self.evaluations += self.swarm_size
        self.update_leaders_density_estimator()

        reference_points = []
        for i, _ in enumerate(self.reference_points):
            point = self.problem.create_solution()
            point.objectives = self.reference_points[i]
            reference_points.append(point)

        observable_data = {'evaluations': self.evaluations,
                           'computing time': self.get_current_computing_time(),
                           'population': self.get_result() + reference_points,
                           'reference_front': self.problem.reference_front}

        self.observable.notify_all(**observable_data)

    def initialize_global_best(self, swarm: List[FloatSolution]) -> None:
        for particle in swarm:
            for leader in self.leaders:
                leader.add(copy(particle))

    def update_global_best(self, swarm: List[FloatSolution]) -> None:
        for particle in swarm:
            for leader in self.leaders:
                leader.add(copy(particle))

    def get_result(self) -> List[FloatSolution]:
        result = []

        for leader in self.leaders:
            for solution in leader.solution_list:
                result.append(solution)

        return result

    def select_global_best(self) -> FloatSolution:
        selected = False
        selected_swarm_index = 0

        while not selected:
            selected_swarm_index = random.randint(0, len(self.leaders) - 1)
            if len(self.leaders[selected_swarm_index].solution_list) != 0:
                selected = True

        leaders = self.leaders[selected_swarm_index].solution_list

        if len(leaders) > 2:
            particles = random.sample(leaders, 2)

            if self.leaders[selected_swarm_index].comparator.compare(particles[0], particles[1]) < 1:
                best_global = copy(particles[0])
            else:
                best_global = copy(particles[1])
        else:
            best_global = copy(self.leaders[selected_swarm_index].solution_list[0])

        return best_global



from copy import copy
from typing import TypeVar, List

from jmetal.component.evaluator import Evaluator
from jmetal.core.algorithm import EvolutionaryAlgorithm
from jmetal.core.operator import Mutation, Crossover, Selection
from jmetal.core.problem import Problem

S = TypeVar('S')
R = TypeVar('R')

"""
.. module:: evolutionary_algorithm
   :platform: Unix, Windows
   :synopsis: Implementation of Evolutionary Algorithms.

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class ElitistEvolutionStrategy(EvolutionaryAlgorithm[S, R]):

    def __init__(self,
                 problem: Problem[S],
                 mu: int,
                 lambd_a: int,
                 max_evaluations: int,
                 mutation: Mutation[S]):
        super(ElitistEvolutionStrategy, self).__init__()
        self.problem = problem
        self.mu = mu
        self.lambd_a = lambd_a
        self.max_evaluations = max_evaluations
        self.mutation = mutation

    def init_progress(self):
        self.evaluations = self.mu

    def update_progress(self):
        self.evaluations += self.lambd_a

    def is_stopping_condition_reached(self) -> bool:
        return self.evaluations >= self.max_evaluations

    def create_initial_population(self) -> List[S]:
        population = []
        for i in range(self.mu):
            population.append(self.problem.create_solution())
        return population

    def evaluate_population(self, population: List[S]) -> List[S]:
        for solution in population:
            self.problem.evaluate(solution)
        return population

    def selection(self, population: List[S]) -> List[S]:
        return population

    def reproduction(self, population: List[S]) -> List[S]:
        offspring_population = []
        for solution in population:
            for j in range(int(self.lambd_a / self.mu)):
                new_solution = copy(solution)
                offspring_population.append(self.mutation.execute(new_solution))

        return offspring_population

    def replacement(self, population: List[S], offspring_population: List[S]) -> List[S]:
        for solution in offspring_population:
            self.population.append(solution)

        population.sort(key=lambda s: s.objectives[0])

        new_population = []
        for i in range(self.mu):
            new_population.append(population[i])

        return new_population

    def get_result(self) -> R:
        return self.population[0]

    def get_name(self) -> str:
        return 'Elitist evolution Strategy'


class NonElitistEvolutionStrategy(ElitistEvolutionStrategy[S, R]):

    def __init__(self,
                 problem: Problem[S],
                 mu: int,
                 lambd_a: int,
                 max_evaluations: int,
                 mutation: Mutation[S]):
        super(NonElitistEvolutionStrategy, self).__init__(problem, mu, lambd_a, max_evaluations, mutation)

    def replacement(self, population: List[S], offspring_population: List[S]) -> List[S]:
        offspring_population.sort(key=lambda s: s.objectives[0])

        new_population = []
        for i in range(self.mu):
            new_population.append(offspring_population[i])

        return new_population

    def get_name(self) -> str:
        return 'Non-Elitist evolution Strategy'


class GenerationalGeneticAlgorithm(EvolutionaryAlgorithm[S, R]):

    def __init__(self,
                 problem: Problem[S],
                 population_size: int,
                 max_evaluations: int,
                 mutation: Mutation[S],
                 crossover: Crossover[S, S],
                 selection: Selection[List[S], S],
                 evaluator: Evaluator[S]):
        super(GenerationalGeneticAlgorithm, self).__init__()
        self.problem = problem
        self.population_size = population_size
        self.max_evaluations = max_evaluations
        self.mutation_operator = mutation
        self.crossover_operator = crossover
        self.selection_operator = selection
        self.evaluator = evaluator

    def init_progress(self):
        self.evaluations = self.population_size

    def update_progress(self):
        self.evaluations += self.population_size

        observable_data = {'evaluations': self.evaluations,
                           'computing time': self.get_current_computing_time(),
                           'population': self.population,
                           'reference_front': self.problem.reference_front}

        self.observable.notify_all(**observable_data)

    def is_stopping_condition_reached(self) -> bool:
        return self.evaluations >= self.max_evaluations

    def create_initial_population(self) -> List[S]:
        population = []

        for i in range(self.population_size):
            population.append(self.problem.create_solution())

        return population

    def evaluate_population(self, population: List[S]):
        return self.evaluator.evaluate(population, self.problem)

    def selection(self, population: List[S]):
        mating_population = []

        for i in range(self.population_size):
            solution = self.selection_operator.execute(self.population)
            mating_population.append(solution)

        return mating_population

    def reproduction(self, population: List[S]) -> List[S]:
        number_of_parents_to_combine = self.crossover_operator.get_number_of_parents()
        self.__check_number_of_parents(number_of_parents_to_combine)

        offspring_population = []
        for i in range(0, self.population_size, number_of_parents_to_combine):
            parents = []
            for j in range(number_of_parents_to_combine):
                parents.append(population[i + j])

            offspring = self.crossover_operator.execute(parents)

            for solution in offspring:
                self.mutation_operator.execute(solution)
                offspring_population.append(solution)

        return offspring_population

    def replacement(self, population: List[S], offspring_population: List[S]) -> List[S]:
        population.sort(key=lambda s: s.objectives[0])

        offspring_population.append(population[0])
        offspring_population.append(population[1])

        offspring_population.sort(key=lambda s: s.objectives[0])

        offspring_population.pop()
        offspring_population.pop()

        return offspring_population

    def get_result(self) -> R:
        """ :return: The best individual of the population.
        """
        return self.population[0]

    def __check_number_of_parents(self, number_of_parents_for_crossover: int):
        if self.population_size % number_of_parents_for_crossover != 0:
            raise Exception('Wrong number of parents')

    def get_name(self) -> str:
        return 'Generational Genetic Algorithm'



from abc import ABCMeta, abstractmethod
from typing import TypeVar, List

from jmetal.component.comparator import DominanceComparator

S = TypeVar('S')


class Ranking(List[S]):

    __metaclass__ = ABCMeta

    def __init__(self):
        super(Ranking, self).__init__()
        self.number_of_comparisons = 0
        self.ranked_sublists = []

    @abstractmethod
    def compute_ranking(self, solution_list: List[S]):
        pass

    def get_subfront(self, rank: int):
        if rank >= len(self.ranked_sublists):
            raise Exception('Invalid rank: {0}. Max rank: {1}'.format(rank, len(self.ranked_sublists) - 1))
        return self.ranked_sublists[rank]

    def get_number_of_subfronts(self):
        return len(self.ranked_sublists)


class FastNonDominatedRanking(Ranking[List[S]]):
    """ Class implementing the non-dominated ranking of NSGA-II. """

    def __init__(self):
        super(FastNonDominatedRanking, self).__init__()

    def compute_ranking(self, solution_list: List[S]):
        # number of solutions dominating solution ith
        dominating_ith = [0 for _ in range(len(solution_list))]

        # list of solutions dominated by solution ith
        ith_dominated = [[] for _ in range(len(solution_list))]

        # front[i] contains the list of solutions belonging to front i
        front = [[] for _ in range(len(solution_list) + 1)]

        for p in range(len(solution_list) - 1):
            for q in range(p + 1, len(solution_list)):
                dominance_test_result = DominanceComparator().compare(solution_list[p], solution_list[q])
                self.number_of_comparisons += 1

                if dominance_test_result == -1:
                    ith_dominated[p].append(q)
                    dominating_ith[q] += 1
                elif dominance_test_result is 1:
                    ith_dominated[q].append(p)
                    dominating_ith[p] += 1

        for i in range(len(solution_list)):
            if dominating_ith[i] is 0:
                front[0].append(i)
                solution_list[i].attributes['dominance_ranking'] = 0

        i = 0
        while len(front[i]) != 0:
            i += 1
            for p in front[i - 1]:
                if p <= len(ith_dominated):
                    for q in ith_dominated[p]:
                        dominating_ith[q] -= 1
                        if dominating_ith[q] is 0:
                            front[i].append(q)
                            solution_list[q].attributes['dominance_ranking'] = i

        self.ranked_sublists = [[]] * i
        for j in range(i):
            q = [0] * len(front[j])
            for k in range(len(front[j])):
                q[k] = solution_list[front[j][k]]
            self.ranked_sublists[j] = q

        return self.ranked_sublists


class EfficientNonDominatedRanking(Ranking[List[S]]):
    """ Class implementing the EDS (efficient non-dominated sorting) algorithm. """

    def __init__(self):
        super(EfficientNonDominatedRanking, self).__init__()

    def compute_ranking(self, solution_list: List[S]):
        # todo
        return self.ranked_sublists



from .crossover import NullCrossover, SBX, SP
from .mutation import NullMutation, BitFlip, Polynomial, IntegerPolynomial, Uniform, SimpleRandom
from .selection import BestSolutionSelection, BinaryTournamentSelection, BinaryTournament2Selection, \
    RandomSolutionSelection, NaryRandomSolutionSelection, RankingAndCrowdingDistanceSelection

__all__ = [
    'NullCrossover', 'SBX', 'SP',
    'NullMutation', 'BitFlip', 'Polynomial', 'IntegerPolynomial', 'Uniform', 'SimpleRandom',
    'BestSolutionSelection', 'BinaryTournamentSelection', 'BinaryTournament2Selection', 'RandomSolutionSelection',
    'NaryRandomSolutionSelection', 'RankingAndCrowdingDistanceSelection'
]



from .graphic import FrontPlot, ScatterStreaming
from .laboratory import Experiment
from .solution_list_output import SolutionList

__all__ = [
    'FrontPlot', 'ScatterStreaming',
    'Experiment',
    'SolutionList'
]



from setuptools import find_packages

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(
    name='jmetalpy',
    version='0.5.1',
    description='JMetalPy. Python version of the jMetal framework',
    author='Antonio J. Nebro',
    author_email='antonio@lcc.uma.es',
    maintainer='Antonio J. Nebro, Antonio Benítez-Hidalgo',
    maintainer_email='antonio@lcc.uma.es, antonio.b@uma.es',
    license='MIT',
    url='https://github.com/jMetal/jMetalPy',
    long_description=open('README.md').read(),
    packages=find_packages(exclude=['test_']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Programming Language :: Python :: 3.6'
    ],
    install_requires=[
        'tqdm',
        'numpy',
        'pandas',
        'plotly',
        'matplotlib==2.0.2',
    ],
    tests_require=[
        'mockito'
        'PyHamcrest',
    ]
)



from math import pi, cos, atan

from jmetal.core.solution import FloatSolution
from jmetal.core.problem import FloatProblem

"""
.. module:: constrained
   :platform: Unix, Windows
   :synopsis: Constrained test problems for multi-objective optimization

.. moduleauthor:: Antonio J. Nebro <antonio@lcc.uma.es>
"""


class Srinivas(FloatProblem):
    """ Class representing problem Srinivas. """

    def __init__(self, rf_path: str=None):
        super(Srinivas, self).__init__(rf_path=rf_path)
        self.number_of_objectives = 2
        self.number_of_variables = 2
        self.number_of_constraints = 2

        self.obj_directions = [self.MINIMIZE, self.MINIMIZE]
        self.obj_labels = ['f(x)', 'f(y)']

        self.lower_bound = [-20.0 for _ in range(self.number_of_variables)]
        self.upper_bound = [20.0 for _ in range(self.number_of_variables)]

        FloatSolution.lower_bound = self.lower_bound
        FloatSolution.upper_bound = self.upper_bound

    def evaluate(self, solution: FloatSolution) -> FloatSolution:
        x1 = solution.variables[0]
        x2 = solution.variables[1]

        solution.objectives[0] = 2.0 + (x1 - 2.0) * (x1 - 2.0) + (x2 - 1.0) * (x2 - 1.0)
        solution.objectives[1] = 9.0 * x1 - (x2 - 1.0) * (x2 - 1.0)

        return solution

    def evaluate_constraints(self, solution: FloatSolution) -> None:
        constraints = [0.0 for _ in range(self.number_of_constraints)]

        x1 = solution.variables[0]
        x2 = solution.variables[1]

        constraints[0] = 1.0 - (x1 * x1 + x2 * x2) / 225.0
        constraints[1] = (3.0 * x2 - x1) / 10.0 - 1.0

        overall_constraint_violation = 0.0
        number_of_violated_constraints = 0.0

        for constrain in constraints:
            if constrain < 0.0:
                overall_constraint_violation += constrain
                number_of_violated_constraints += 1

        solution.attributes['overall_constraint_violation'] = overall_constraint_violation
        solution.attributes['number_of_violated_constraints'] = number_of_violated_constraints

    def get_name(self):
        return 'Srinivas'


class Tanaka(FloatProblem):
    """ Class representing problem Tanaka """

    def __init__(self, rf_path: str=None):
        super(Tanaka, self).__init__(rf_path=rf_path)
        self.number_of_objectives = 2
        self.number_of_variables = 2
        self.number_of_constraints = 2

        self.obj_directions = [self.MINIMIZE, self.MINIMIZE]
        self.obj_labels = ['f(x)', 'f(y)']

        self.lower_bound = [10e-5 for _ in range(self.number_of_variables)]
        self.upper_bound = [pi for _ in range(self.number_of_variables)]

        FloatSolution.lower_bound = self.lower_bound
        FloatSolution.upper_bound = self.upper_bound

    def evaluate(self, solution: FloatSolution) -> FloatSolution:
        solution.objectives[0] = solution.variables[0]
        solution.objectives[1] = solution.variables[1]

        return solution

    def evaluate_constraints(self, solution: FloatSolution) -> None:
        constraints = [0.0 for _ in range(self.number_of_constraints)]

        x1 = solution.variables[0]
        x2 = solution.variables[1]

        constraints[0] = (x1 * x1 + x2 * x2 - 1.0 - 0.1 * cos(16.0 * atan(x1 / x2)))
        constraints[1] = -2.0 * ((x1 - 0.5) * (x1 - 0.5) + (x2 - 0.5) * (x2 - 0.5) - 0.5)

        overall_constraint_violation = 0.0
        number_of_violated_constraints = 0.0

        for constrain in constraints:
            if constrain < 0.0:
                overall_constraint_violation += constrain
                number_of_violated_constraints += 1

        solution.attributes['overall_constraint_violation'] = overall_constraint_violation
        solution.attributes['number_of_violated_constraints'] = number_of_violated_constraints

    def get_name(self):
        return 'Tanaka'



import logging
from abc import ABCMeta
from string import Template
from typing import TypeVar, List

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from plotly import graph_objs as go
from plotly.offline import plot
from pandas import DataFrame, merge

jMetalPyLogger = logging.getLogger('jMetalPy')
S = TypeVar('S')

"""
.. module:: visualization
   :platform: Unix, Windows
   :synopsis: Classes for plotting fronts.

.. moduleauthor:: Antonio Benítez-Hidalgo <antonio.b@uma.es>
"""


class Plot:

    __metaclass__ = ABCMeta

    def __init__(self, plot_title: str, axis_labels: list):
        self.plot_title = plot_title
        self.axis_labels = axis_labels

        self.number_of_objectives: int = None

    @staticmethod
    def get_objectives(front: List[S]) -> DataFrame:
        """ Get objectives for each solution of the front.

        :param front: List of solutions.
        :return: Pandas dataframe with one column for each objective and one row for each solution. """
        if front is None:
            raise Exception('Front is none!')

        return DataFrame(list(solution.objectives for solution in front))


class ScatterStreaming(Plot):

    def __init__(self, plot_title: str, axis_labels: list = None):
        """ Creates a new :class:`ScatterStreaming` instance. Suitable for problems with 2 or 3 objectives in streaming.

        :param plot_title: Title of the diagram.
        :param axis_labels: List of axis labels. """
        super(ScatterStreaming, self).__init__(plot_title, axis_labels)

        import warnings
        warnings.filterwarnings("ignore", ".*GUI is implemented.*")

        self.fig = plt.figure()
        self.sc = None
        self.axis = None

    def plot(self, front: List[S], reference_front: List[S], filename: str = '', show: bool = True) -> None:
        """ Plot a front of solutions (2D or 3D).

        :param front: List of solutions.
        :param reference_front: Reference solution list (if any).
        :param filename: If specified, save the plot into a file.
        :param show: If True, show the final diagram (default to True). """
        objectives = self.get_objectives(front)

        # Initialize plot
        self.number_of_objectives = objectives.shape[1]
        self.__initialize()

        if reference_front:
            jMetalPyLogger.info('Reference front found')
            ref_objectives = self.get_objectives(reference_front)

            if self.number_of_objectives == 2:
                self.__plot(ref_objectives[0], ref_objectives[1], None,
                            color='#323232', marker='*', markersize=3)
            else:
                self.__plot(ref_objectives[0], ref_objectives[1], ref_objectives[2],
                            color='#323232', marker='*', markersize=3)

        if self.number_of_objectives == 2:
            self.__plot(objectives[0], objectives[1], None, color='#98FB98', marker='o', markersize=3)
        else:
            self.__plot(objectives[0], objectives[1], objectives[2], color='#98FB98', marker='o', markersize=3)

        if filename:
            self.fig.savefig(filename, format='png', dpi=200)
        if show:
            self.fig.canvas.mpl_connect('pick_event', lambda event: self.__pick_handler(front, event))
            plt.show()

    def update(self, front: List[S], reference_front: List[S], rename_title: str = '',
               persistence: bool = True) -> None:
        """ Update an already created plot.

        :param front: List of solutions.
        :param reference_front: Reference solution list (if any).
        :param rename_title: New title of the plot.
        :param persistence: If True, keep old points; else, replace them with new values.
        """
        if self.sc is None:
            jMetalPyLogger.warning('Plot must be initialized first.')
            self.plot(front, reference_front, show=False)
            return

        objectives = self.get_objectives(front)

        if persistence:
            # Replace with new points
            self.sc.set_data(objectives[0], objectives[1])

            if self.number_of_objectives == 3:
                self.sc.set_3d_properties(objectives[2])
        else:
            # Add new points
            if self.number_of_objectives == 2:
                self.__plot(objectives[0], objectives[1], None, color='#98FB98', marker='o', markersize=3)
            else:
                self.__plot(objectives[0], objectives[1], objectives[2], color='#98FB98', marker='o', markersize=3)

        # Also, add event handler
        event_handler = \
            self.fig.canvas.mpl_connect('pick_event', lambda event: self.__pick_handler(front, event))

        # Update title with new times and evaluations
        self.fig.suptitle(rename_title, fontsize=13)

        # Re-align the axis
        self.axis.relim()
        self.axis.autoscale_view(True, True, True)

        try:
            # Draw
            self.fig.canvas.draw()
        except KeyboardInterrupt:
            pass

        plt.pause(0.01)

        # Disconnect the pick event for the next update
        self.fig.canvas.mpl_disconnect(event_handler)

    def __initialize(self) -> None:
        """ Initialize the scatter plot for the first time. """
        jMetalPyLogger.info('Generating plot')

        # Initialize a plot
        self.fig.canvas.set_window_title('jMetalPy')

        if self.number_of_objectives == 2:
            self.axis = self.fig.add_subplot(111)

            # Stylize axis
            self.axis.spines['top'].set_visible(False)
            self.axis.spines['right'].set_visible(False)
            self.axis.get_xaxis().tick_bottom()
            self.axis.get_yaxis().tick_left()
        elif self.number_of_objectives == 3:
            self.axis = Axes3D(self.fig)
            self.axis.autoscale(enable=True, axis='both')
        else:
            raise Exception('Number of objectives must be either 2 or 3')

        self.axis.set_autoscale_on(True)
        self.axis.autoscale_view(True, True, True)

        # Style options
        self.axis.grid(color='#f0f0f5', linestyle='-', linewidth=1, alpha=0.5)
        self.fig.suptitle(self.plot_title, fontsize=13)

        jMetalPyLogger.info('Plot initialized')

    def __plot(self, x_values, y_values, z_values, **kwargs) -> None:
        if self.number_of_objectives == 2:
            self.sc, = self.axis.plot(x_values, y_values, ls='None', picker=10, **kwargs)
        else:
            self.sc, = self.axis.plot(x_values, y_values, z_values, ls='None', picker=10, **kwargs)

    def __pick_handler(self, front: List[S], event):
        """ Handler for picking points from the plot. """
        line, ind = event.artist, event.ind[0]
        x, y = line.get_xdata(), line.get_ydata()

        jMetalPyLogger.debug('Selected front point ({0}): ({1}, {2})'.format(ind, x[ind], y[ind]))

        sol = next((solution for solution in front
                    if solution.objectives[0] == x[ind] and solution.objectives[1] == y[ind]), None)

        if sol is not None:
            with open('{0}-{1}'.format(x[ind], y[ind]), 'w') as of:
                of.write(sol.__str__())
        else:
            jMetalPyLogger.warning('Solution is none')
            return True


class FrontPlot(Plot):

    def __init__(self, plot_title: str, axis_labels: list = None):
        """ Creates a new :class:`FrontPlot` instance. Suitable for problems with 2 or more objectives.

        :param plot_title: Title of the graph.
        :param axis_labels: List of axis labels. """
        super(FrontPlot, self).__init__(plot_title, axis_labels)

        self.figure: go.Figure = None
        self.layout = None
        self.data = []

    def plot(self, front: List[S], reference_front: List[S] = None, normalize: bool = False) -> None:
        """ Plot a front of solutions (2D, 3D or parallel coordinates).

        :param front: List of solutions.
        :param reference_front: Reference solution list (if any).
        :param normalize: Normalize the input front between 0 and 1 (for problems with more than 3 objectives). """
        self.__initialize()

        if reference_front:
            objectives = self.get_objectives(reference_front)
            trace = self.__generate_trace(objectives=objectives, legend='reference front', normalize=normalize,
                                          color='rgb(2, 130, 242)')
            self.data.append(trace)

        objectives = self.get_objectives(front)
        metadata = list(solution.__str__() for solution in front)
        trace = self.__generate_trace(objectives=objectives, metadata=metadata, legend='front', normalize=normalize,
                                      symbol='diamond-open')
        self.data.append(trace)

        self.figure = go.Figure(data=self.data, layout=self.layout)

    def update(self, data: List[S], normalize: bool = False, legend: str = '') -> None:
        """ Update an already created graph with new data.

        :param data: List of solutions to be included.
        :param legend: Legend to be included.
        :param normalize: Normalize the input front between 0 and 1 (for problems with more than 3 objectives). """
        if self.figure is None:
            jMetalPyLogger.warning('Plot must be initialized first.')
            self.plot(data, reference_front=None, normalize=normalize)
            return

        objectives = self.get_objectives(data)
        new_data = self.__generate_trace(objectives=objectives, legend=legend, normalize=normalize,
                                         color='rgb(255, 170, 0)')
        self.data.append(new_data)

        self.figure = go.Figure(data=self.data, layout=self.layout)

    def to_html(self, filename: str='front') -> str:
        """ Export the graph to an interactive HTML (solutions can be selected to show some metadata).

        :param filename: Output file name.
        :return: Script as string. """
        html_string = '''
        <!DOCTYPE html>
        <html>
            <head>
                <meta charset="utf-8"/>
                <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
                <script src="https://unpkg.com/sweetalert2@7.7.0/dist/sweetalert2.all.js"></script>
                <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.1/css/bootstrap.min.css">
                <style>
                    .float{
                      position:fixed;
                      right:40px;
                      bottom:40px;
                    }
                </style>
            </head>
            <body>
                <a class="float" href="https://jmetalpy.readthedocs.io/en/latest/">
                  <img src="https://raw.githubusercontent.com/jMetal/jMetalPy/master/docs/source/jmetalpy.png" height="20px"/>
                </a>
                ''' + self.export(include_plotlyjs=False) + '''
                <script>                
                    var myPlot = document.querySelectorAll('div')[0];
                    myPlot.on('plotly_click', function(data){
                        var pts = '';
                        
                        for(var i=0; i < data.points.length; i++){
                            pts = '(x, y) = ('+data.points[i].x +', '+ data.points[i].y.toPrecision(4)+')';
                            cs = data.points[i].customdata
                        }
                        
                        if(typeof cs !== "undefined"){
                            swal({
                              title: 'Closest solution clicked:',
                              text: cs,
                              type: 'info',
                              position: 'bottom-end'
                            })
                        }
                    });
                    
                    window.onresize = function() {
                       Plotly.Plots.resize(myPlot);
                    };
                </script>
            </body>
        </html>'''

        with open(filename + '.html', 'w') as outf:
            outf.write(html_string)

        return html_string

    def export(self, filename: str = '', include_plotlyjs: bool = False) -> str:
        """ Export as a `div` for embedding the graph in an HTML file.

        :param filename: Output file name (if desired, default to None).
        :param include_plotlyjs: If True, include plot.ly JS script (default to False).
        :return: Script as string. """
        script = plot(self.figure, output_type='div', include_plotlyjs=include_plotlyjs, show_link=False)

        if filename:
            with open(filename + '.html', 'w') as outf:
                outf.write(script)

        return script

    def __initialize(self):
        """ Initialize the graph for the first time. """
        jMetalPyLogger.info('Generating graph')

        self.layout = go.Layout(
            margin=dict(l=80, r=80, b=80, t=150),
            height=800,
            title=self.plot_title,
            scene=dict(
                xaxis=dict(title=self.axis_labels[0:1][0] if self.axis_labels[0:1] else None),
                yaxis=dict(title=self.axis_labels[1:2][0] if self.axis_labels[1:2] else None),
                zaxis=dict(title=self.axis_labels[2:3][0] if self.axis_labels[2:3] else None)
            ),
            hovermode='closest'
        )

    def __generate_trace(self, objectives: DataFrame, metadata: list = None, legend: str = '', normalize: bool = False,
                         **kwargs):
        number_of_objectives = objectives.shape[1]

        if normalize:
            objectives = (objectives - objectives.min()) / (objectives.max() - objectives.min())

        marker = dict(
            color='rgb(127, 127, 127)',
            size=3,
            symbol='x',
            line=dict(
                color='rgb(204, 204, 204)',
                width=1
            ),
            opacity=0.8
        )
        marker.update(**kwargs)

        if number_of_objectives == 2:
            trace = go.Scattergl(
                x=objectives[0],
                y=objectives[1],
                mode='markers',
                marker=marker,
                name=legend,
                customdata=metadata
            )
        elif number_of_objectives == 3:
            trace = go.Scatter3d(
                x=objectives[0],
                y=objectives[1],
                z=objectives[2],
                mode='markers',
                marker=marker,
                name=legend,
                customdata=metadata
            )
        else:
            dimensions = list()
            for column in objectives:
                dimensions.append(
                    dict(range=[0, 1],
                         label=self.axis_labels[column:column+1][0] if self.axis_labels[column:column+1] else None,
                         values=objectives[column])
                )

            trace = go.Parcoords(
                line=dict(color='blue'),
                dimensions=dimensions,
                name=legend,
            )

        return trace

    def __save(self, filename: str = 'front', show: bool = False) -> None:
        """ Save the graph. """
        plot(self.figure, filename=filename + '.html', auto_open=show, show_link=False)



from os.path import dirname, join
import unittest

from jmetal.core.solution import Solution
from jmetal.problem import ZDT1
from jmetal.component.quality_indicator import HyperVolume


class HyperVolumeTestCases(unittest.TestCase):

    def setUp(self):
        self.file_path = dirname(join(dirname(__file__)))

    def test_should_hypervolume_return_5_0(self):
        reference_point = [2, 2, 2]

        solution1 = Solution(1, 3)
        solution1.objectives = [1, 0, 1]

        solution2 = Solution(1, 3)
        solution2.objectives = [0, 1, 0]

        front = [solution1, solution2]

        hv = HyperVolume(reference_point)
        value = hv.compute(front)

        self.assertEqual(5.0, value)

    def test_should_hypervolume_return_the_correct_value_when_applied_to_the_ZDT1_reference_front(self):
        problem = ZDT1(rf_path='resources/reference_front/ZDT1.pf')
        reference_point = [1, 1]

        hv = HyperVolume(reference_point)
        value = hv.compute(problem.reference_front)

        self.assertAlmostEqual(0.666, value, delta=0.001)


if __name__ == '__main__':
    unittest.main()
